@@ -6,13 +6,14 @@ import tempfile
 import weakref
 from datetime import timedelta, datetime
 from threading import get_ident
+from urllib.parse import urlparse
 
 import dask
 from distributed import Client
 from distributed.security import Security
 from distributed.utils import LoopRunner, sync, thread_state, format_bytes, log_errors
 from tornado import gen
-from tornado.httpclient import AsyncHTTPClient, HTTPRequest
+from tornado.httpclient import AsyncHTTPClient, HTTPRequest, HTTPError
 from tornado.httputil import HTTPHeaders
 
 from .cookiejar import CookieJar
@@ -24,7 +25,18 @@ from .options import Options
 
 del comm
 
-__all__ = ("Gateway", "GatewayCluster")
+__all__ = ("Gateway", "GatewayCluster", "GatewayClusterError", "GatewayServerError")
+
+
+class GatewayClusterError(Exception):
+    """Exception related to starting/stopping/scaling of a gateway cluster"""
+
+
+class GatewayServerError(Exception):
+    """Exception related to the operation of the gateway server.
+
+    Indicates an internal error in the gateway server.
+    """
 
 
 class GatewaySecurity(Security):
@@ -56,10 +68,10 @@ class GatewaySecurity(Security):
             ctx.verify_mode = ssl.CERT_REQUIRED
             ctx.check_hostname = False
             ctx.load_cert_chain(cert_path, key_path)
-        return {"ssl_context": ctx}
+        return {"ssl_context": ctx, "require_encryption": True}
 
 
-class _EnumMixin(object):
+class _IntEnum(enum.IntEnum):
     @classmethod
     def _create(cls, x):
         return x if isinstance(x, cls) else cls.from_name(x)
@@ -74,7 +86,7 @@ class _EnumMixin(object):
         raise ValueError("%r is not a valid %s" % (name, cls.__name__))
 
 
-class ClusterStatus(_EnumMixin, enum.IntEnum):
+class ClusterStatus(_IntEnum):
     """The cluster's status.
 
     Attributes
@@ -115,8 +127,8 @@ class ClusterReport(object):
         The cluster's status.
     scheduler_address : str or None
         The address of the scheduler, or None if not currently running.
-    dashboard_address : str or None
-        The address of the dashboard, or None if not currently running.
+    dashboard_link : str or None
+        A link to the dashboard, or None if not currently running.
     start_time : datetime.datetime
         The time the cluster was started.
     stop_time : datetime.datetime or None
@@ -134,7 +146,7 @@ class ClusterReport(object):
         "options",
         "status",
         "scheduler_address",
-        "dashboard_address",
+        "dashboard_link",
         "start_time",
         "stop_time",
         "tls_cert",
@@ -147,7 +159,7 @@ class ClusterReport(object):
         options,
         status,
         scheduler_address,
-        dashboard_address,
+        dashboard_link,
         start_time,
         stop_time,
         tls_cert=None,
@@ -155,9 +167,9 @@ class ClusterReport(object):
     ):
         self.name = name
         self.options = options
-        self.status = ClusterStatus._create(status)
+        self.status = status
         self.scheduler_address = scheduler_address
-        self.dashboard_address = dashboard_address
+        self.dashboard_link = dashboard_link
         self.start_time = start_time
         self.stop_time = stop_time
         self.tls_cert = tls_cert
@@ -166,22 +178,40 @@ class ClusterReport(object):
     def __repr__(self):
         return "ClusterReport<name=%s, status=%s>" % (self.name, self.status.name)
 
+    @property
+    def security(self):
+        """A security object for this cluster, if present, else None"""
+        return (
+            None
+            if self.tls_key is None
+            else GatewaySecurity(tls_key=self.tls_key, tls_cert=self.tls_cert)
+        )
+
     @classmethod
-    def _from_json(cls, gateway_address, msg):
+    def _from_json(cls, gateway_address, proxy_address, msg):
         start_time = datetime.fromtimestamp(msg.pop("start_time") / 1000)
         stop_time = msg.pop("stop_time")
         if stop_time is not None:
             stop_time = datetime.fromtimestamp(stop_time / 1000)
 
+        name = msg.pop("name")
+        status = ClusterStatus._create(msg.pop("status"))
+        scheduler_address = (
+            "%s/%s" % (proxy_address, name) if status == ClusterStatus.RUNNING else None
+        )
+
         dashboard_route = msg.pop("dashboard_route")
-        dashboard_address = (
-            gateway_address + dashboard_route if dashboard_route else None
+        dashboard_link = (
+            "%s%s" % (gateway_address, dashboard_route) if dashboard_route else None
         )
 
         return cls(
             start_time=start_time,
             stop_time=stop_time,
-            dashboard_address=dashboard_address,
+            name=name,
+            status=status,
+            scheduler_address=scheduler_address,
+            dashboard_link=dashboard_link,
             **msg,
         )
 
@@ -193,6 +223,10 @@ class Gateway(object):
     ----------
     address : str, optional
         The address to the gateway server.
+    proxy_address : str, int, optional
+        The address of the scheduler proxy server. If an int, it's used as the
+        port, with the host/ip taken from ``address``. Provide a full address
+        if a different host/ip should be used.
     auth : GatewayAuth, optional
         The authentication method to use.
     asynchronous : bool, optional
@@ -203,14 +237,34 @@ class Gateway(object):
         asynchronous mode, otherwise a background loop is started.
     """
 
-    def __init__(self, address=None, auth=None, asynchronous=False, loop=None):
+    def __init__(
+        self, address=None, proxy_address=None, auth=None, asynchronous=False, loop=None
+    ):
         if address is None:
             address = dask.config.get("gateway.address")
         if address is None:
             raise ValueError(
                 "No dask-gateway address provided or found in configuration"
             )
-        self.address = address.rstrip("/")
+        address = address.rstrip("/")
+
+        if proxy_address is None:
+            proxy_address = dask.config.get("gateway.proxy-address")
+        if proxy_address is None:
+            raise ValueError(
+                "No dask-gateway proxy address provided or found in configuration"
+            )
+        if isinstance(proxy_address, int):
+            parsed = urlparse(address)
+            proxy_netloc = "%s:%d" % (parsed.hostname, proxy_address)
+        elif isinstance(proxy_address, str):
+            parsed = urlparse(proxy_address)
+            proxy_netloc = parsed.netloc if parsed.netloc else proxy_address
+        proxy_address = "gateway://%s" % proxy_netloc
+
+        self.address = address
+        self.proxy_address = proxy_address
+
         self._auth = get_auth(auth)
         self._cookie_jar = CookieJar()
 
@@ -279,20 +333,37 @@ class Gateway(object):
             return sync(self.loop, func, *args, **kwargs)
 
     async def _fetch(self, req):
-        self._cookie_jar.pre_request(req)
-        resp = await self._http_client.fetch(req, raise_error=False)
-        if resp.code == 401:
-            context = self._auth.pre_request(req, resp)
+        try:
+            self._cookie_jar.pre_request(req)
             resp = await self._http_client.fetch(req, raise_error=False)
-            self._auth.post_response(req, resp, context)
-        self._cookie_jar.post_response(resp)
-        if resp.error:
-            if resp.code in {404, 409, 422}:
-                raise ValueError(json.loads(resp.body)["error"])
-            elif resp.code == 599:
+            if resp.code == 401:
+                context = self._auth.pre_request(req, resp)
+                resp = await self._http_client.fetch(req, raise_error=False)
+                self._auth.post_response(req, resp, context)
+            self._cookie_jar.post_response(resp)
+            if resp.error:
+                if resp.code == 599:
+                    raise TimeoutError("Request timed out")
+                else:
+                    try:
+                        msg = json.loads(resp.body)["error"]
+                    except Exception:
+                        msg = resp.body.decode()
+
+                    if resp.code in {404, 422}:
+                        raise ValueError(msg)
+                    elif resp.code == 409:
+                        raise GatewayClusterError(msg)
+                    elif resp.code == 500:
+                        raise GatewayServerError(msg)
+                    else:
+                        resp.rethrow()
+        except HTTPError as exc:
+            # Tornado 6 still raises these above with raise_error=False
+            if exc.code == 599:
                 raise TimeoutError("Request timed out")
-            else:
-                resp.rethrow()
+            # Should never get here!
+            raise
         return resp
 
     async def _clusters(self, status=None):
@@ -309,7 +380,7 @@ class Gateway(object):
         req = HTTPRequest(url=url)
         resp = await self._fetch(req)
         return [
-            ClusterReport._from_json(self.address, r)
+            ClusterReport._from_json(self.address, self.proxy_address, r)
             for r in json.loads(resp.body).values()
         ]
 
@@ -397,7 +468,9 @@ class Gateway(object):
         url = "%s/gateway/api/clusters/%s%s" % (self.address, cluster_name, params)
         req = HTTPRequest(url=url)
         resp = await self._fetch(req)
-        return ClusterReport._from_json(self.address, json.loads(resp.body))
+        return ClusterReport._from_json(
+            self.address, self.proxy_address, json.loads(resp.body)
+        )
 
     async def _connect(self, cluster_name):
         while True:
@@ -408,14 +481,22 @@ class Gateway(object):
                 pass
             else:
                 if report.status is ClusterStatus.RUNNING:
-                    return GatewayCluster(self, report)
+                    return GatewayCluster(
+                        gateway=self,
+                        name=report.name,
+                        scheduler_address=report.scheduler_address,
+                        dashboard_link=report.dashboard_link,
+                        security=report.security,
+                    )
                 elif report.status is ClusterStatus.FAILED:
-                    raise Exception(
+                    raise GatewayClusterError(
                         "Cluster %r failed to start, see logs for "
                         "more information" % cluster_name
                     )
                 elif report.status is ClusterStatus.STOPPED:
-                    raise Exception("Cluster %r is already stopped" % cluster_name)
+                    raise GatewayClusterError(
+                        "Cluster %r is already stopped" % cluster_name
+                    )
             # Not started yet, try again later
             await gen.sleep(0.5)
 
@@ -432,6 +513,8 @@ class Gateway(object):
         cluster_name = await self._submit(**kwargs)
         try:
             return await self._connect(cluster_name)
+        except GatewayClusterError:
+            raise
         except BaseException:
             # Ensure cluster is stopped on error
             await self._stop_cluster(cluster_name)
@@ -527,14 +610,12 @@ class GatewayCluster(object):
     calling this class directly.
     """
 
-    def __init__(self, gateway, report):
+    def __init__(self, gateway, name, scheduler_address, dashboard_link, security):
         self._gateway = gateway
-        self.name = report.name
-        self.scheduler_address = report.scheduler_address
-        self.dashboard_address = report.dashboard_address
-        self.security = GatewaySecurity(
-            tls_key=report.tls_key, tls_cert=report.tls_cert
-        )
+        self.name = name
+        self.scheduler_address = scheduler_address
+        self.dashboard_link = dashboard_link
+        self.security = security
         self._internal_client = None
         # XXX: Ensure client is closed. Currently disconnected clients try to
         # reconnect forever, which spams the schedulerproxy. Once this bug is
@@ -658,11 +739,10 @@ class GatewayCluster(object):
 
         elements = [title, HBox([status, request, scale])]
 
-        if self.dashboard_address is not None:
-            link_address = "%s/status" % self.dashboard_address
+        if self.dashboard_link is not None:
             link = HTML(
-                '<p><b>Dashboard: </b><a href="%s" target="_blank">%s'
-                "</a></p>\n" % (link_address, link_address)
+                '<p><b>Dashboard: </b><a href="{0}" target="_blank">{0}'
+                "</a></p>\n".format(self.dashboard_link)
             )
             elements.append(link)
 
@@ -676,4 +756,26 @@ class GatewayCluster(object):
         widget = self._widget()
         if widget is not None:
             return widget._ipython_display_(**kwargs)
-        print(self)
+        else:
+            from IPython.display import display
+
+            data = {"text/plain": repr(self), "text/html": self._repr_html_()}
+            display(data, raw=True)
+
+    def _repr_html_(self):
+        if self.dashboard_link is not None:
+            dashboard = "<a href='{0}' target='_blank'>{0}</a>".format(
+                self.dashboard_link
+            )
+        else:
+            dashboard = "Not Available"
+        return (
+            "<div style='background-color: #f2f2f2; display: inline-block; "
+            "padding: 10px; border: 1px solid #999999;'>\n"
+            "  <h3>GatewayCluster</h3>\n"
+            "  <ul>\n"
+            "    <li><b>Name: </b>{name}\n"
+            "    <li><b>Dashboard: </b>{dashboard}\n"
+            "  </ul>\n"
+            "</div>\n"
+        ).format(name=self.name, dashboard=dashboard)
